//
//  FeedListViewController.swift
//  ReviewsFeed
//
//  Created by Dmitrii Ivanov on 21/07/2020.
//  Copyright © 2020 ING. All rights reserved.
//

import Combine
import Foundation
import ReviewsFoundationKit
import ReviewsUIKit
import SwiftUI
import UIKit

public class FeedListViewController: UITableViewController {

    // MARK: Constants
    private let viewModel: ViewModel
    
    // MARK: Properties
    private var cancellables: Set<AnyCancellable> = []
    
    // MARK: Outlets
    private lazy var filterButton = {
        UIBarButtonItem(
            title: NSLocalizedString(
                .Key.Navigation.Button.filter,
                bundle: .module,
                comment: .empty
            ),
            image: .Icon.filter,
            primaryAction: nil,
            menu: .init(
                title: NSLocalizedString(
                    .Key.Menu.filter,
                    bundle: .module,
                    comment: .empty
                ),
                image: UIImage.Icon.star,
                children: {
                    FilterOption.allCases.map { option -> UIAction in
                        .init(
                            title: option.text,
                            image: .init(systemName: option.icon)
                        ) { [weak self] _ in
                            self?.viewModel.filter(by: option)
                        }
                    }
                }()
            )
        )
    }()

    // MARK: Initialisers
    public init(configuration: Configuration = .init()) {
        self.viewModel = .init(configuration: configuration)

        super.init(style: .plain)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    // MARK: Computed
    var items: [Review] {
        viewModel.items
    }

    // MARK: UIViewController
    public override func viewDidLoad() {
        super.viewDidLoad()

        setNavigationBar()
        registerTableCells()
        bindViewModel()
        
        viewModel.fetch()
    }
    
    // MARK: UITableViewDataSource
    public override func tableView(
        _ tableView: UITableView,
        numberOfRowsInSection section: Int
    ) -> Int {
        viewModel.itemsCount
    }
    
    public override func tableView(
        _ tableView: UITableView,
        cellForRowAt indexPath: IndexPath
    ) -> UITableViewCell {
        if viewModel.isWordsShowing && indexPath.row == 0 {
            makeTopWordsCell(tableView)
        } else {
            makeFeedItemCell(tableView, at: indexPath.row)
        }
    }
    
    // MARK: UITableViewDelegate
    public override func tableView(
        _ tableView: UITableView,
        didSelectRowAt indexPath: IndexPath
    ) {
        let details = FeedItemViewController(items[indexPath.row])
        
        tableView.deselectRow(
            at: indexPath,
            animated: true
        )

        navigationController?.pushViewController(details, animated: true)
    }
    
}

// MARK: - Helpers
private extension FeedListViewController {
    
    // MARK: Functions
    func bindViewModel() {
        viewModel.$filter
            .receive(on: RunLoop.main)
            .sink { [weak self] option in
                self?.updateFilterMenu(option)
                self?.tableView.reloadData()
            }
            .store(in: &cancellables)

        viewModel.$isFilterEnabled
            .removeDuplicates()
            .receive(on: RunLoop.main)
            .sink { [weak self] enabled in
                self?.filterButton.isEnabled = enabled
            }
            .store(in: &cancellables)
        
        viewModel.$isLoading
            .dropFirst()
            .filter { $0 == false }
            .removeDuplicates()
            .receive(on: RunLoop.main)
            .sink { [weak self] _ in
                self?.tableView.reloadData()
            }
            .store(in: &cancellables)
    }
    
    func makeFeedItemCell(
        _ tableView: UITableView,
        at row: Int
    ) -> UITableViewCell {
        guard
            let cell = FeedItemCell.dequeue(from: tableView),
            let item = viewModel.item(for: row)
        else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    FeedItemView(item)
                }
            } else {
                HostingConfiguration {
                    FeedItemView(item)
                }
            }
        }()
        
        return cell
    }
    
    func makeTopWordsCell(_ tableView: UITableView) -> UITableViewCell {
        guard let cell = TopWordsCell.dequeue(from: tableView) else {
            return .init()
        }
        
        cell.separatorInset = .init(width: tableView.bounds.size.width)
        cell.contentConfiguration = {
            if #available(iOS 16.0, *) {
                UIHostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            } else {
                HostingConfiguration {
                    TopWordsView(viewModel.words)
                }
            }
        }()

        return cell
    }
    
    func registerTableCells() {
        FeedItemCell.register(in: tableView)
        TopWordsCell.register(in: tableView)
    }
    
    func setNavigationBar() {
        navigationController?.navigationBar.prefersLargeTitles = true
        navigationController?.navigationBar.isTranslucent = true

        navigationItem.rightBarButtonItem = filterButton
        navigationItem.title = NSLocalizedString(
            .Key.Navigation.title,
            bundle: .module,
            comment: .empty
        )
    }
    
    func updateFilterMenu(_ option: FilterOption) {
        filterButton
            .menu?
            .children
            .compactMap { $0 as? UIAction }
            .forEach { action in
                action.state = action.title == option.text
                    ? .on
                    : .off
            }
    }
    
}

// MARK: - Configuration
extension FeedListViewController {
    public struct Configuration {
        
        // MARK: Constants
        let appID: String
        let countryCode: String
        let session: URLSessionConfiguration
        
        // MARK: Initialisers
        public init(
            appID: String = "474495017",
            countryCode: String = "nl",
            session: URLSessionConfiguration = .ephemeral
        ) {
            self.appID = appID
            self.countryCode = countryCode
            self.session = session
        }
        
    }
}

// MARK: - String+Constants
private extension String {
    enum Cell {
        static let feedItem = "FeedItemCell"
    }
    
    enum Key {
        enum Menu {
            static let filter = "common.filter.menu.title.text"
        }

        enum Navigation {
            static let title = "view.feed-list.navigation-bar.title.text"
            
            enum Button {
                static let filter = "view.feed-list.navigation-bar.button.filter-list.text"
            }
        }
    }
}

// MARK: - UIEdgeInsets+Inits
private extension UIEdgeInsets {
    
    // MARK: Initialisers
    init(width: CGFloat) {
        self = .init(
            top: 0,
            left: width,
            bottom: 0,
            right: 0
        )
    }
    
}

// MARK: - Previews
#if DEBUG
import ReviewsiTunesKit

@available(iOS 17.0, *)
#Preview("Feed List loading reviews") {
    MockURLProtocol.response = .init(statusCode: 200)
    
    return UINavigationController(
        rootViewController: FeedListViewController(configuration: .init(session: .mock))
    )
}

@available(iOS 17.0, *)
#Preview("Feed List with few reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: [
            .init(
                id: 1,
                author: "Some author name #1 here",
                title: "Some review title #1 goes here...",
                content: "Some long, explanatory review comment #1 goes here...",
                rating: 3,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 2,
                author: "Some author name #2 here",
                title: "Some review title #2 goes here...",
                content: "Some long, explanatory review comment #2 goes here...",
                rating: 5,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 3,
                author: "Some author name #3 here",
                title: "Some review title #3 goes here...",
                content: "Some long, explanatory review comment #3 goes here...",
                rating: 1,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 4,
                author: "Some author name #4 here",
                title: "Some review title #4 goes here...",
                content: "Some long, explanatory review comment #4 goes here...",
                rating: 4,
                version: "v1.0.0",
                updated: .init()
            ),
            .init(
                id: 5,
                author: "Some author name #5 here",
                title: "Some review title #5 goes here...",
                content: "Some long, explanatory review comment #5 goes here...",
                rating: 2,
                version: "v1.0.0",
                updated: .init()
            ),
        ])
    )
    
    return UINavigationController(
        rootViewController: FeedListViewController(configuration: .init(session: .mock))
    )
}

@available(iOS 17.0, *)
#Preview("Feed List with no reviews") {
    MockURLProtocol.response = .init(
        statusCode: 200,
        object: Feed(entries: [])
    )
    
    return UINavigationController(
        rootViewController: FeedListViewController(configuration: .init(session: .mock))
    )
}

@available(iOS 17.0, *)
#Preview("Feed List with live reviews") {
    UINavigationController(rootViewController: FeedListViewController())
}
#endif
